/************************************
 * Steven Stanek
 * January 13, 2001
 * YZiForm Implementation (yziform.c)
 * Version: 0.0.1d
 ************************************/
 
#include "yziform.h"
#include "yziformgzwrap.h"
#include <string.h>

#define  BEGINNING_DEFS	16
#define  BEGINNING_STRS 256
#define	 BEGIN_TAG	"<Begin YZiForm Image"
#define	 DEF_TAG	"<Definitions>"
#define  IMAGE_TAG	"<Image"
#define	 TERM_TAG	"<End Image>"
#define	 MAX_CH_VAL	256*256
#define  READ_BUF_SIZE 1024
#define  READ_INCREMENT 16384
#define	 RESET_READ		64

#if macintosh
	#undef 	 	feof
#endif

typedef struct
{
	color_t color;
	char *name;
	
} definition_t;

struct yzImageData
{
	definition_t **defs;
	char *image;
	int width, height, numdefs;
	void *fastdraw;
	color_t temp_color; //For use with Local Bitmaps
};

//Read in Strings Functions Prototypes
static char* getBtwTags(FILE* dFile);
static char* extractDefStr(char **str);

//Read in Colors Function Prototypes
static void getColors(char *definitions, yzImage dImage);
static definition_t readDefinition(char **str);
static void enterDefinition(definition_t aDef, yzImage dImage);
static int getNumLineBreaks(char *str);
static int hashString(char *str, int high);

//Drawing Image Helper Function Prototypes
static color_t* findColor(char *name, yzImage dImage);
static char * getNextToken(char **str);

//For Fast Draw various bit depths
static inline void enableFD16(yzImage aImage);
static inline void enableFD32(yzImage aImage);
static void drawWithFD(yzImage aImage, int x, int y);

//For Bitmaps in ImageSegment
static void hexsToColor(char *hexs, color_t *color);
static inline int fourHexesToInt(char *hexes);
static inline int hexCharToInt(char hex);


//Public Stuff
yzImage openImageFile(FILE *toOpen)
{
	char buffer[128], *definitions;
	int width, height;
	yzImage dImage;
	
	assert(dImage = malloc(sizeof(struct yzImageData)));
	dImage->fastdraw = 0;
	
	//Read & Check Begin Tag / Compression Support
	fgets(buffer, 127, toOpen);
	
	#ifdef yziformgzwrap_h_
		if(!strncmp(buffer, COMP_IMAGE_STR, strlen(COMP_IMAGE_STR)))
		{
			free(dImage);
			return readCompressedFileImage(toOpen);
		}
	#endif
	
	assert(!strncmp(buffer, BEGIN_TAG, strlen(BEGIN_TAG)));
	
	//Read & Check Definitions Tag
	fgets(buffer, 127, toOpen);
	assert(!strncmp(buffer, DEF_TAG, strlen(DEF_TAG)));
	
	//Read Definitions
	definitions = getBtwTags(toOpen);
	
	//Read & Check Image Tag
	fgets(buffer, 127, toOpen);
	assert(!strncmp(buffer, IMAGE_TAG, strlen(IMAGE_TAG)));
	sscanf(buffer, "<Image %d,%d>", &width, &height);
	
	//Read Image
	dImage->image = getBtwTags(toOpen);
	
	//Read & Check Image Termination Tag
	fgets(buffer, 127, toOpen);
	assert(!strncmp(buffer, TERM_TAG, strlen(TERM_TAG)));
	
	//Configure Image
	dImage->width = width;
	dImage->height = height;
	getColors(definitions, dImage);
	
	free(definitions);	
	return dImage;
}


yzImage getImageFromString(char *str)
{
	char *defs;
	yzImage dImage;

	assert(dImage = malloc(sizeof(struct yzImageData)));
	dImage->fastdraw = 0;
	
	//Read & Check Begin Tag / Compression Support
	#ifdef yziformgzwrap_h_
		if(!strncmp(str, COMP_IMAGE_STR, strlen(COMP_IMAGE_STR)))
		{
			free(dImage);
			return processCompressedString(str);//Most likely won't work
		}
	#endif
	
	assert(!strncmp(str, BEGIN_TAG, strlen(BEGIN_TAG)));
	assert(str = strchr(str, '\n'));
	str++;
	
	//Read & Check Definitions Tag
	assert(!strncmp(str, DEF_TAG, strlen(DEF_TAG)));
	assert(str = strchr(str, '\n')); 
	str++;
	
	//Read Definitions
	defs = extractDefStr(&str);
	
	//Deal with image tag
	assert(!strncmp(str, IMAGE_TAG, strlen(IMAGE_TAG)));
	sscanf(str, "<Image %d,%d>", &(dImage->width), &(dImage->height));
	assert(str = strchr(str, '\n')); 
	str++;
	
	//Read in image string
	dImage->image = extractDefStr(&str);
	
	//Check Termination
	assert(!strncmp(str, TERM_TAG, strlen(TERM_TAG)));
	
	getColors(defs, dImage);
	free(defs);
	
	return dImage;
}


void drawImage(yzImage toDraw, int x, int y)
{
	int endx, endy, nx, ny;
	char *image_str, *next_token;
	color_t *nextColor;
	
	if(toDraw->fastdraw)
	{
		drawWithFD(toDraw, x, y);
		return;
	}
	
	image_str = toDraw->image;
	endx = x + toDraw->width;
	endy = y + toDraw->height;
	nx = x;
	ny = y;
	
	for(; y < endy; y++)
	{
		for(x=nx; x < endx; x++)
		{
			next_token = getNextToken(&image_str);
			if(strcmp(next_token, TRANS_STR))
			{
				nextColor = findColor(next_token, toDraw);
				yzPixel(x, y, *nextColor);
			}
				
			free(next_token);
		}
		image_str++;
	}
}


color_t *getColorAt(yzImage aImage, int x, int y)
{
	int i, j;
	char *image_str, *nextToken;

	if(x >= getImageWidth(aImage) || y >= getImageHeight(aImage) || x < 0 || y < 0)
		return 0; 
	
	for(i=0; i < y; i++)
	{
		for(j=0; j < getImageWidth(aImage); j++)
		{
			nextToken  = getNextToken(&image_str);
			
			if(i == y -1 && j - 1 == x)
				return findColor(nextToken, aImage);
		}
		
		image_str ++;
	}
	
	return 0;
}


color_t** toColors(yzImage toConvert)
{
	int i,j;
	char *image_str,* next_token;
	color_t ** dColors;
	
	image_str = toConvert->image;
	assert(dColors = malloc(toConvert->height * sizeof(color_t*)));
	
	for(i=0; i < toConvert->height; i++)
	{
		assert(dColors[i]=malloc(toConvert->width * sizeof(color_t)));
		
		for(j=0;j < toConvert->width; j++)
		{
			next_token = getNextToken(&image_str);
			
			if(!strcmp(next_token, TRANS_STR))
			{
				dColors[i][j].r=dColors[i][j].g=dColors[i][j].b=65535;
				dColors[i][j].a=TRANS_A;
			}
			
			else
				dColors[i][j] = *findColor(next_token, toConvert);
				
			free(next_token);
		}
		image_str++;
	}
	
	return dColors;
}


void destroyImage(yzImage toKill)
{
	int i;
		
	free(toKill->image);
	
	for(i=0; i < toKill->numdefs; i++)
	{
		free(toKill->defs[i]->name);
		free(toKill->defs[i]);	
	}
	
	if(toKill->fastdraw)
		free(toKill->fastdraw);
	
	free(toKill->defs);
	free(toKill);	
}


int getImageWidth(yzImage aImage)
{
	return aImage->width;
}


int getImageHeight(yzImage aImage)
{
	return aImage->height;
}


void enableFastDraw(yzImage aImage)
{
	int depth;
	
	depth = yzGetDepth();
	
	if(yzGetDepth() == 16)
		enableFD16(aImage);
	else if(yzGetDepth() == 32)
		enableFD32(aImage);
}

void* getFastDrawArray(yzImage aImage)
{
	return aImage->fastdraw;
}


//Private Stuff
//NOTE: getDefString, is used only for file I/O
//Reads data between tags
static char* getBtwTags(FILE* dFile)
{
	char *indata, *temp, nextchar;
	int alloced, startpos, amountread;
	
	startpos = ftell(dFile);
	
	assert(indata = malloc(READ_INCREMENT * sizeof(char)));
	alloced = READ_INCREMENT;
	
	while(1)
	{
		amountread = ftell(dFile) - startpos;
		
		if(amountread >= (alloced - RESET_READ))
		{
			alloced += READ_INCREMENT;
			
			assert(temp = malloc(sizeof(char) * alloced));
			strcpy(temp, indata);
			free(indata);
			indata = temp;
		}
		
		if(amountread != 0)
			fgets(indata + amountread - 1, alloced - amountread - 1, dFile);
		else
			fgets(indata, alloced - 1, dFile);
		
		nextchar = fgetc(dFile);
		ungetc(nextchar, dFile);
		
		if(nextchar == '<')
			break;
	}
	
	return indata;	
}

//Extracts Definitions Str from char array and creates and returns copy
//Used only for locally stored strings
static char* extractDefStr(char **str)
{
	char *defs, *end;
	
	end = strchr(*str, '<');
	assert(defs = malloc(sizeof(char) * (end - *str +1)));
	strncpy(defs, *str, end - *str);
	defs[end-*str] = '\0';
	*str = end;
	
	return defs;
}


/* Function: static void getColors(char *definitions, yzImage dImage)
 * This function takes a string containing all definitions and 
 * compiles them into the given yzImage. It does this by iterating
 * through the definition lines, extracting one line and parsing
 * it, and getting the name and color values.
 */
static void getColors(char *definitions, yzImage dImage)
{
	int i;
	
	dImage->numdefs = getNumLineBreaks(definitions);
	assert(dImage->defs = malloc(dImage->numdefs * sizeof(definition_t*)));
	
	for(i=0; i < dImage->numdefs; i++)
		dImage->defs[i] = 0;
	
	for(i=0; 1; i++)
	{	
		if(definitions[0] == '\0')
			break;
		
		enterDefinition(readDefinition(&definitions), dImage);
	}
	assert(i == dImage->numdefs);
}


/* Function: static definition_t readDefinition(char **str)
 * This function reads a definition from a String and advances
 * the string accordingly.
 */
static definition_t readDefinition(char **str)
{
	int r, g, b;
	char name[128], temp[128];
	definition_t def;
	
	sscanf(*str, "%s = %d,%d,%d", name, &r,&g,&b);
	(*str)++; //Eliminate \n	
		
	if(name[0] == '\"')
	{
		strcpy(temp, name+1);
		temp[strlen(temp)-1] = '\0';
		strcpy(name, temp);
	}
		
	assert(r < MAX_CH_VAL && g < MAX_CH_VAL && b < MAX_CH_VAL);
	
	assert(def.name = malloc(strlen(name) * sizeof(char)));
	strcpy(def.name, name);
	def.color.r = r;
	def.color.g = g;
	def.color.b = b;
	def.color.a = TRANS_A + 1;
	
	*str = strchr(*str, '\n')+1; //10==\n on Mac. All YZiForm files use 10 for \n
	
	return def;
}


/* Function: static void enterDefinition(definition_t aDef, yzImage dImage)
 * This function is tasked with adding definitions to the hashTable in
 * the passed in yzImage. It uses a rolling method of adding them. IE: Rather
 * than extending into linked lists, it looks for the next open bucket.
 * Like everything else in this library, it has error checking.
 */
static void enterDefinition(definition_t aDef, yzImage dImage)
{
	int code, i;
	
	code = hashString(aDef.name, dImage->numdefs);

	for(i = 0; i < dImage->numdefs; i++)
	{
		if(dImage->defs[code] == 0)
		{
			assert(dImage->defs[code] = malloc(sizeof(definition_t)));
			*(dImage->defs[code]) = aDef;
			return;
		}
		
		code = (code+1) % dImage->numdefs;
	}
	
	printf("Num Buckets is wrong\n");
	exit(5);
}


/* Function: static int getNumLineBreaks(char *str)
 * This function returns the number of '\n' characters
 * in a string.
 */
static int getNumLineBreaks(char *str)
{
	int numBreaks=0;
	char presChar;
	
	while(1)
	{
		presChar = str[0];
		
		if(presChar == '\0')
			break;
		
		if(presChar == '\n') //Always uses 10 as \n
			numBreaks++;
		str++;
	}
	return numBreaks;
}


/* Function: static int hashString(char *str, int high)
 * Takes a string and creates a numeric representation
 * of it. This number's upper limit is defined by the 
 * variable high.
 */
static int hashString(char *str, int high)
{
	int hashnum =1;
	
	while(1)
	{
		if(str[0] == '\0')
			break;
			
		hashnum += (str[0]*hashnum);
		str++;
	}
	
	hashnum *= 314159;
	
	return abs(hashnum % high);
}



/* Function: static color_t* findColor(char *name, yzImage dImage)
 * Returns pointer to color_t which corresponds to the respective
 * name.
 */
static color_t* findColor(char *name, yzImage dImage)
{
	int i, code;
	definition_t *presDef;
	
	if(name[0] == BITMAP_ID)
	{
		name ++;
		hexsToColor(name, &(dImage->temp_color));
		return &(dImage->temp_color);
	}
	
	code = hashString(name, dImage->numdefs);
	
	for(i=0; i < dImage->numdefs; i++)
	{
		presDef = dImage->defs[code];
		if(presDef && strcmp(presDef->name, name)==0)
			return &(presDef->color);
		code = (code +1) % dImage->numdefs;
	}
	
	printf("ErrorIndexed color doesn't exist");
	exit(0);
	
	return 0;
}


/* Function: static char * getNextToken(char **str)
 * Returns a dynamically allocated string containing the name of the
 * next color in the string and advances the str pointer accordingly
 */
static char * getNextToken(char **str)
{
	char *toreturn;
	int length;
	
	if(*str[0] == '\"')
	{
		(*str)++;
		length = strchr(*str, '\"')- *str;
		assert(toreturn = malloc((length +1) * sizeof(char)));
		strncpy(toreturn, *str, length);
		toreturn[length] = '\0';
		(*str)+= length +1;
	}
	
	else
	{
		assert(toreturn = malloc(2 * sizeof(char)));
		toreturn[0] = *str[0];
		toreturn[1] = '\0';
		(*str)++;
	}
	
	return toreturn;
}


//Fast Draw Stuff
static inline void enableFD16(yzImage aImage)
{
	int i, j;
	char *image_str, *presToken;
	unsigned short *data, r, g, b;
	color_t *presColor;
	
	if(sizeof(unsigned short)*8 != 16)
		return;
	
	data = malloc(2 * getImageHeight(aImage) * getImageWidth(aImage));
	
	if(!data)
		return;
		
	image_str = aImage->image;
	
	for(i=0; i < getImageHeight(aImage); i++)
	{
		for(j=0; j < getImageWidth(aImage); j++)
		{
			presToken = getNextToken(&image_str);
			presColor = findColor(presToken,aImage);
			free(presToken);
			
			if(!presColor)
				return;
			
			r = presColor->r / 2048;
			g = presColor->g / 2048;
			b = presColor->b / 2048;
			
			data[j + i * getImageWidth(aImage)] = (r<<10) | (g<<5) | (b);
		}
		image_str++;
	}
	aImage->fastdraw = data;
}


static inline void enableFD32(yzImage aImage)
{
	int i, j;
	char *image_str, *presToken;
	unsigned long *data, r, g, b;
	color_t *presColor;
	
	if(sizeof(unsigned long)*8 != 32)
		return;
	
	data = malloc(4 * getImageHeight(aImage) * getImageWidth(aImage));
	
	if(!data)
		return;
		
	image_str = aImage->image;
	
	for(i=0; i < getImageHeight(aImage); i++)
	{
		for(j=0; j < getImageWidth(aImage); j++)
		{
			presToken = getNextToken(&image_str);
			presColor = findColor(presToken,aImage);
			free(presToken);
			
			r = presColor->r / 256;
			g = presColor->g / 256;
			b = presColor->b / 256;
			
			data[j + i * getImageWidth(aImage)] = (r<<16) | (g<<8) | (b);
		}
		image_str ++;
	}
	
	aImage->fastdraw = data;
}


static void drawWithFD(yzImage aImage, int x, int y)
{
	if(!aImage->fastdraw)
		return;
	
	yzDisplayBits(x, y, getImageWidth(aImage), getImageHeight(aImage), 
		getImageWidth(aImage) * yzGetDepth() / 8, aImage->fastdraw);
}


static void hexsToColor(char *hexs, color_t *color)
{	
	color->r = fourHexesToInt(hexs);
	hexs+=4;
	color->g = fourHexesToInt(hexs);
	hexs+=4;
	color->b = fourHexesToInt(hexs);
	color->a = 0;	
}


static inline int fourHexesToInt(char *hexes)
{
	int final;
	
	final = hexCharToInt(hexes[0]) * 4096 + hexCharToInt(hexes[1]) * 256 + 
			hexCharToInt(hexes[2]) * 16 + hexCharToInt(hexes[3]);
			
	return final;
}


static inline  int hexCharToInt(char hex)
{
	if(hex >='0' && hex <= '9')
		return hex - '0';
	else if(hex >= 'A' && hex <= 'F')
		return hex - 'A' + 10;
	
	printf("Error Not Hex Values");
	exit(0);
	return 0;
}